/*
 * dldLPUAlgorithm.cxx
 *
 */

#include "dldLPUAlgorithm.h"

namespace dld
{

/************************ Constructor *******************************/
LPUAlgorithm::LPUAlgorithm()
{
	P.clear();
	sigma.clear();
	lambda.clear();
	model.clear();
	risks.clear();
	pValues = "";
	sValues = "";
	lambdaValues = "";
	output = "";
	results = "";
	groundtruth = "";
	positivedata = "";
	unlabeleddata = "";

	num_positive = 0;
	num_labeled = 0;

	optimizeModel = false;
}

/************************ destructor *******************************/
LPUAlgorithm::~LPUAlgorithm()
{
	FileWriter::deleteFile(std::string(MERGE));
}

/************************ setPositiveDataFile *******************************/
void LPUAlgorithm::setPositiveDataFile(const std::string p)
{
	positivedata = p;
	num_positive = FileReader::countLines( positivedata );
}

/************************ setUnsampledDataFile *******************************/
void LPUAlgorithm::setUnlabeledDataFile(const std::string u)
{
	unlabeleddata = u;
	num_labeled = FileReader::countLines( unlabeleddata );
}

/************************ setModelFile *******************************/
void LPUAlgorithm::setModelFile(const std::string m)
{
	model = vtksys::SystemTools::GetFilenameWithoutLastExtension(m);
	model_ext = vtksys::SystemTools::GetFilenameLastExtension(m);
}

/************************ setOutputFile *******************************/
void LPUAlgorithm::setOutputFile(const std::string o)
{
	output = vtksys::SystemTools::GetFilenameWithoutLastExtension(o);
	output_ext = vtksys::SystemTools::GetFilenameLastExtension(o);
}

/***************************** doAll *******************************/
void LPUAlgorithm::doAll(void)
{
	if (num_labeled <= 0 || num_positive <= 0)
	{
		std::cerr << "Error on sample data: Program will fail" << std::endl;
	}
	else
	{
		P = FileReader::loadVectorFromFile<float>(pValues);
		lambda = FileReader::loadVectorFromFile<float>(lambdaValues);
		sigma = FileReader::loadVectorFromFile<float>(sValues);
		FileWriter::mergeFiles(positivedata,unlabeleddata,std::string(MERGE));

		Trainer* t = new Trainer();
		t->setTrainPath(path);
		t->setTrainFile(std::string(MERGE));
		Predictor* p = new Predictor();
		p->setTestPath(path);
		p->setTestFile(std::string(MERGE));

		unsigned int total_samples = num_positive + num_labeled;
		float ratio = (float) num_labeled / (float) num_positive;

		for (unsigned int i = 0; i < P.size(); ++i) //test all proportions
		{
			float w_one = 2*P[i]*ratio;
			float best_risk = 1e7; //Check this up
			unsigned int ind_l = 0, ind_s = 0;
			for (unsigned int j = 0; j < lambda.size(); ++j) //all lambdas
			{
				float C = 1 / ((2*lambda[j])*(1+2*P[i])*total_samples);
				for (unsigned int k = 0; k < sigma.size(); ++k) // all sigmas
				{
					std::string model_it = model + "_p" + Utils::to_string<int>(i)
						+ "_l" + Utils::to_string<int>(j) + "_s"
						+ Utils::to_string<int>(k) + model_ext;
					t->train(model_it,C,w_one,1,sigma[k]);
					std::string output_it = output + "_p" + Utils::to_string<int>(i)
						+ "_l" + Utils::to_string<int>(j) + "_s"
						+ Utils::to_string<int>(k) + output_ext;
					p->predict(model_it,output_it);
					float risk = evaluateRisk(output_it,P[i]);
					if (optimizeModel)
					{
						if (best_risk > risk)
						{
							best_risk = risk;
							ind_l = j;
							ind_s = k;
						}
					}
					else	//If no optimization: print out values
						std::cout << "P: " << P[i] << " lambda: " << lambda[j]
									<< " sigma: " << sigma[k] << " risk: "
									<< risk << std::endl;
				}
			}
			if (optimizeModel)
			{
				risks.push_back(best_risk); //extension
				std::cout << "Optimized P: " << P[i] << " lambda: "
						<< lambda[ind_l] << " sigma: " << sigma[ind_s]
						<< " risk: " << best_risk << std::endl;
			}
		}

		delete t;
		delete p;
	}
}

/***************************** trainLPU *******************************/
void LPUAlgorithm::trainLPU(void) const
{
	Trainer* t = new Trainer();
	t->setTrainFile(std::string(MERGE));

	unsigned int total_samples = num_positive + num_labeled;
	float ratio = (float) num_labeled / (float) num_positive;

	for (unsigned int i = 0; i < P.size(); ++i) //test all proportions
	{
		float w_one = 2*P[i]*ratio;
		for (unsigned int j = 0; j < lambda.size(); ++j) //all lambdas
		{
			float C = 1 / ((2*lambda[j])*(1+2*P[i])*total_samples);
			for (unsigned int k = 0; k < sigma.size(); ++k) // all sigmas
			{
				std::string model_it = model + "_p" + Utils::to_string<int>(i)
						+ "_l" + Utils::to_string<int>(j) + "_s"
						+ Utils::to_string<int>(k);

				t->train(model_it,C,w_one,1,sigma[k]);
			}
		}
	}

	delete t;
}

/***************************** predictLPU *******************************/
void LPUAlgorithm::predictLPU(void) const
{
	//TODO Single call to predict
}


/***************************** evaluateRisk *******************************/
float LPUAlgorithm::evaluateRisk(const std::string o, float p) const
{
	std::vector<int> all = FileReader::loadVectorFromFile<int>(o.c_str());
	unsigned int term_one = 0;
	unsigned int term_two = 0;

	for (unsigned int i = 0; i < num_positive; ++i)
	{
		if (all[i] <= 0) ++term_one;
	}

	for (unsigned int i = num_positive; i < num_positive + num_labeled; ++i)
	{
		if (all[i] > 0) ++term_two;
	}

	return ((2*p*term_one/(num_positive*(1+2*p)))) +
						(term_two/(num_labeled*2*p));
}

} // end namespace

